"""Stream type classes for tap-fabdb."""

from typing import Any, Dict, Final, Optional

import requests
from singer_sdk import typing as th
from singer_sdk.streams import RESTStream


class FabDbStream(RESTStream):
    """FaB DB base stream class."""

    url_base = "https://api.fabdb.net"

    def get_url_params(self, context: Optional[dict], next_page_token: Optional[int]) -> Dict[str, Any]:
        """Override the RESTStream get_url_params method to support FaB DB pagination."""
        params = {
            "per_page": 100,
        }
        if next_page_token is not None:
            params["page"] = next_page_token
        return params

    def get_next_page_token(self, response: requests.Response, previous_token: Optional[int]) -> Optional[int]:
        """Override the RESTStream get_next_page_token method to support FaB DB pagination."""
        current_page = previous_token or response.json().get("current_page", 1)
        if response.json()["links"].get("next") is not None:
            return current_page + 1
        else:
            return None


class CardsStream(FabDbStream):
    """Cards stream representing FaB DB Cards."""

    name = "cards"
    path = "/cards"
    primary_keys = ["identifier"]
    records_jsonpath = "$.data[*]"
    stats = {"cost", "attack", "defense", "resource"}
    stats_properties = [th.Property(stat, th.StringType) for stat in stats]

    schema: Final = th.PropertiesList(
        th.Property("identifier", th.StringType, required=True),
        th.Property("name", th.StringType, required=True),
        th.Property("rarity", th.StringType, required=True),
        th.Property("keywords", th.ArrayType(th.StringType), required=True),
        th.Property("stats", th.ObjectType(*stats_properties), required=True),
        th.Property("text", th.StringType),
        th.Property("image", th.StringType),
        th.Property("totalSideboard", th.IntegerType),
        th.Property("printings", th.ArrayType(th.ObjectType())),
        th.Property("flavour", th.StringType),
        th.Property("comments", th.StringType),
        th.Property("banned", th.ArrayType(th.StringType)),
        th.Property("next", th.StringType),
        th.Property("prev", th.StringType),
    ).to_dict()

    def post_process(self, row: dict, context: Optional[dict] = None) -> dict:
        """As needed, append or transform raw data to match expected structure.

        Optional. This method gives developers an opportunity to "clean up" the results
        prior to returning records to the downstream tap - for instance: cleaning,
        renaming, or appending properties to the raw record result returned from the
        API.
        """
        if isinstance(row["stats"], list):
            row["stats"] = row["stats"].pop() if row["stats"] else {}

        for stat in self.stats:
            if stat in row["stats"]:
                row["stats"][stat] = str(row["stats"][stat])
        return row



"""Tests standard tap features using the built-in SDK tests library."""

from unittest.mock import patch

import pytest
import responses
from singer_sdk.testing import get_standard_tap_tests

from tap_fabdb.streams import CardsStream
from tap_fabdb.tap import TapFabDb


class TestFabDbStream:
    """Test specifics of the FabDbStream base class."""

    @pytest.fixture
    def stream(self):
        """Stream fixture."""
        return CardsStream(tap=TapFabDb(), name="cards")

    @patch("tap_fabdb.streams.requests.Response")
    def test_get_next_page_token(self, mock_response, stream):
        """Test the overriden get_next_page_token functionality."""
        mock_response.json.return_value = {"links": {"next": "https://api.fabdb.net/cards?page=2"}}
        next_page_token = stream.get_next_page_token(mock_response, None)
        assert next_page_token == 2
        mock_response.json.return_value = {"links": {"next": None}}
        next_page_token = stream.get_next_page_token(mock_response, 2)
        assert next_page_token is None

    def test_get_url_params(self, stream):
        """Test the overriden get_url_params functionality."""
        params = stream.get_url_params(context={}, next_page_token=2)
        assert params["per_page"] == 100
        assert params["page"] == 2

    def test_post_process(self, stream):
        """Test the post-processing functionality as it relates to Card stats."""
        empty_stats = {"identifier": "foo", "stats": []}
        processed = stream.post_process(empty_stats)
        assert processed["stats"] == {}

        normal_stats = {"identifier": "foo", "stats": {"cost": "1", "defense": "1"}}
        processed = stream.post_process(normal_stats)
        assert processed["stats"] == {"cost": "1", "defense": "1"}


@responses.activate
def test_standard_tap_tests():
    """Run standard tap tests from the SDK."""
    responses.add(
        responses.GET,
        url="https://api.fabdb.net/cards",
        status=200,
        json={
            "data": [
                {
                    "identifier": "foo",
                    "name": "foo",
                    "stats": {"cost": "1", "defense": "1", "attack": "1", "resource": "1"},
                }
            ]
        },
    )
    tests = get_standard_tap_tests(TapFabDb)
    for test in tests:
        test()
